import io
import contextlib
from typing import Optional
import logging
import traceback
import sys
import yattag
import uuid

LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s: %(message)s'


def capture_output(
    target: object,
    args: Optional[tuple] = None,
        kwargs: Optional[dict] = None) -> str:
    """Redirect stdout and stderr into string buffer and capture it.

    target object is executed with optional args, kwargs and all stdout/
    stderr that is captured, returned in string form.

    Args:
        target: object to execute, usually a function.
        args: target positional arguments (default: {None})
        kwargs: target keyword arguments (default: {None})
    """
    if not args:
        args = ()
    if not kwargs:
        kwargs = {}

    with io.StringIO() as sio:
        with contextlib.redirect_stdout(sio), contextlib.redirect_stderr(sio):
            target(*args, **kwargs)
            output = sio.getvalue()
            return output


def setup_logger(
        log_level: str = logging.WARNING, fmt: str = LOG_FORMAT) -> None:
    """Set up basic logging configuration."""
    logging.basicConfig(
        format=LOG_FORMAT,
        level=log_level)


def format_list_to_html(
        line_height=1, collapse_cfg: Optional[dict] = None) -> str:
    """Format list of strings into HTML string.

    Lines are converted into HTML paragraphs. line-height attribute is
    set for all paragraphs. This is default format.

    Optionally can specify collapse_cfg to make part of text
    "togglable" (aka read more/read less). Two ways are supported:
        - bootrstrap collapse (default).
        - attr_toggle on div that is "togglable".

    In both cases max_lines key is required.

    First case is used if max_lines is specified (maximum lines to
    show initially), but not attr_toggle. Optionally collapse_id can be
    passed to use it as anchor for bootsrap collapse toggle. Otherwise
    str(uuid.uuid4()) value is used.

    Second case is used if attr_toggle is used. This way only specified
    custom attribute is added on div that wraps paragraphs that should
    be hidden initially. Paragraphs toggle implementation must be done
    externally in this case.

    Args:
        line_height: paragraph height (default: {1})
        collapse_cfg: toggle show/hide part of text config
            (default: {None}) Used keys:
              - max_lines (int): number of paragraphs to show initially.
              - collapse_id (str): anchor for collapse div
                identification. If not set, will use randomly generated
                ID. Bootstrap implementation only.
              - attr_toggle (tuple): attribute used to toggle show/hide
                part of text. Custom implementation only.

    Returns:
        HTML string
    """
    def build_lines(lines):
        for line in lines:
            # Assuming that line is one paragraph, so `\n` is not needed.
            line = line.replace('\n', '')
            with tag('p'):
                text(line)

    def build_bootstrap_collapse(lines_to_hide):
        # Default to random ID if none was specified. It has
        # very low chance to run in collision, so there should
        # be no problem.
        collapse_id = (
            collapse_cfg.get('collapse_id') or str(uuid.uuid4()))
        with tag('div', id=collapse_id, klass='collapse'):
            build_lines(lines_to_hide)
        # Add button to toggle hidden lines.
        with tag(
            'a',
            ('data-toggle', 'collapse'),
            ('data-target', '#%s' % collapse_id),
                klass='btn btn-link'):
            text('Toggle More')

    def format_html(lines):
        with tag('div'):
            doc.attr(style='line-height: %s' % line_height)
            max_lines = collapse_cfg['max_lines']
            if max_lines == -1:  # Everything is showed.
                build_lines(lines)
            else:
                # Split into lines to show and to hide.
                # Lines to hide.
                lines_to_show = lines[0:max_lines]
                lines_to_hide = lines[max_lines:]
                # Build lines that will be visible all the time.
                build_lines(lines_to_show)
                # Build lines that will be hidden initially.
                if collapse_cfg.get('attr_toggle'):
                    # Using custom specified attribute that should be
                    # used to handle toggling of lines_to_hide (
                    # implementation must be done externally from this
                    # method).
                    with tag('div', collapse_cfg['attr_toggle']):
                        build_lines(lines_to_hide)
                else:
                    # Default to bootstrap implementation.
                    build_bootstrap_collapse(lines_to_hide)
        return doc.getvalue()

    if not collapse_cfg:
        collapse_cfg = {'max_lines': -1}
    doc, tag, text = yattag.Doc().tagtext()
    return format_html


def _format_exception() -> list:
    exc_type, exc_value, exc_traceback = sys.exc_info()
    return traceback.format_exception(exc_type, exc_value, exc_traceback)


def _get_formatted_exception(
        exc_lines: list, formatter=None) -> str:
    if not formatter:
        return ''.join(exc_lines)
    return formatter(exc_lines)


def get_formatted_exception(formatter=None) -> str:
    """Convert exception lines into formatter string.

    How string is formatted, depends on formatter function passed.
    Formatter acts as constructor, so it needs to be executed where its
    closure function can take specified arguments (if there are any) and
    do actual formatting.

    Args:
        formatter: logic how to format (default: {None}). If not
        specified will default to ''.join(exc_lines).

    Returns:
        formatted exception lines string
    """
    exc_lines = _format_exception()
    return _get_formatted_exception(exc_lines, formatter=formatter)



import time
import logging
from typing import Optional


class _TimeIt(object):
    """Decorator  class to track function execution time.

    This decorator should probably be used as convenience. If you need
    to benchmark your code, you should probably use standard module
    timeit.
    """

    _time_map = {
        'ms': ('millisecond(s)', 1000),
        's': ('second(s)', 1),
        'm': ('minute(s)', 1/60),
        'h': ('hour(s)', 1/3600),
    }

    def __init__(
        self,
        log_level: Optional[str] = None,
        msg_fmt: str = "'{_f_name}' took {_dur:.2f} {_uom_name}",
            uom: str = 's') -> None:
        """Initialize specified arguments for _TimeIt object.

        Args:
            log_level: logging level to use (default: {None})
            msg_fmt: message format. Possible keys for message format:
                - main:
                  - '_f_name': function name.
                  - '_dur': time it took to execute function (in format
                    specified in time_fmt).
                  - '_uom_name': unit of measure name (second(s),
                    minute(s) etc.).
                - positional arguments: arg0, arg1, arg2 etc. Will name
                    positional arguments (if any) of function that uses
                    this decorator. If function has __self__ attribute,
                    it will be used for arg0.
                - keyword arguments: keyword arguments (if any) of
                    function that uses this decorator.
                (default: {"'{f_name}' took {dur:.2f} {uom_name}"})
            uom: time unit of measure to output. Possible values:
                'ms' (milliseconds), 's' (seconds), 'm' (minutes), 'h'
                (hours) (default: {'s'})
        """
        self.log_level = log_level
        self.msg_fmt = msg_fmt
        self.uom = uom

    def _get_message_vals(self, func, duration, args=None, kwargs=None):

        def insert_args(args, func, vals):
            pos = 0
            if hasattr(func, '__self__'):
                vals['arg0'] = func.__self__
                # Move position to right, because arg0 is now already
                # assigned.
                pos = 1
            for i, arg in enumerate(args, pos):
                vals['arg%s' % i] = arg

        if not args:
            args = ()
        if not kwargs:
            kwargs = {}
        uom_name, uom_ratio = self._time_map[self.uom]
        # Define main vals.
        vals = {
            '_f_name': func.__name__,
            '_dur': duration * uom_ratio,
            '_uom_name': uom_name
        }
        insert_args(args, func, vals)
        # Insert kwargs. Will overwrite old keys if some already exist.
        vals.update(kwargs)
        return vals

    def _get_message(
        self,
        func,
        duration: float,
        args: Optional[tuple] = None,
            kwargs: Optional[dict] = None):
        vals = self._get_message_vals(func, duration, args=args, kwargs=kwargs)
        return self.msg_fmt.format(**vals)

    def output(self, func, duration, args=None, kwargs=None):
        msg = self._get_message(func, duration, args=args, kwargs=kwargs)
        if self.log_level:
            log_level = self.log_level.lower()
            getattr(logging, log_level)(msg)
        else:
            print(msg)

    def __call__(self, func):
        """Wrap function and track time it took to execute it."""
        def wrapped(*args, **kwargs):
            ts = time.time()
            res = func(*args, **kwargs)
            te = time.time()
            self.output(func, te-ts, args=args, kwargs=kwargs)
            return res
        return wrapped


time_it = _TimeIt
